/*
  ADIFOR visualization library
  (precision-dependent part)

  Vitaly Shmatikov, Summer 1994
*/

#include <limits.h>
#include "PREC_advis_pd_include.h"

/* # of nonzeroes in the l.h.s. of each call in the slice */
static int nz_array[MAX_SLICE_SIZE];
/* Counter of calls in the slice */
static int call_counter = 0;

/* Unique numeric id generator */
static int vector_id = 0;

/* Auxiliary functions */
int nz_compare _P((void *, void *));

/***************************************************************
*                                                              *
*   record_vector identifies the vector in the history file.   *
*                                                              *
***************************************************************/

void record_vector(name, is_tracked, v)
char *name;
short is_tracked;
PREC_vectorstruct *v;
{
#ifdef _VTRACE

 v->id = name; v->is_named = TRUE;
 v->num_id = ++vector_id; v->is_tracked = is_tracked;

 if(global_tracking)
   /* Identify v in the history file */
   fprintf(HF, "@ %s %d\n", v->id, v->num_id);

#endif /* _VTRACE */
}

/*******************************************************
*                                                      *
*     account_vector marks unnamed vector as not       *
*     being tracked and calls record_vector.           *
*                                                      *
*******************************************************/

void account_vector(v)
PREC_vectorstruct *v;
{
#ifdef _VTRACE
 char *name;

 if (v->is_named)
    return;

 name = (char*)malloc(STRING_MAX*sizeof(char));
 sprintf(name, "%p", v);
 record_vector(name, FALSE, v);
#endif /* _VTRACE */
}

/*************************************************************
*                                                            *
*     track_vector markes the vector as tracked and          *
*     calls record_vector.                                   *
*                                                            *
*************************************************************/

void track_vector(v, f_s, index)
PREC_vectorstruct **v;
char *f_s;  /* Pointer to a symbolic name */
int index[];
{
#ifdef _VTRACE
 char *name = (char*)malloc(STRING_MAX*sizeof(char));
 char *inds = (char*)malloc(STRING_MAX*sizeof(char));
 int *ind_pt = index;   /* Pointer to the array of indices */
 char *s;
 unsigned char IndexEmpty = TRUE;

 if((v == NULL) || ((*v) == NULL)) {
   adzero(v);
 }
 
 /* Convert symbolic name from Fortran to C */
 s = f_to_c_str(f_s);

 /* Append indices to the symbolic name of the vector */
 strcpy(name, s);
 if(ind_pt != NULL) {
    /* Got an array of indices */
    if((*ind_pt) != 0) { 
      sprintf(inds, "%d", (*ind_pt)); strcat(name, inds); 
      ind_pt++;
      strcat(name, "("); IndexEmpty = FALSE;
    }
    while((*ind_pt) != 0) { 
      strcat(name, ",");
      sprintf(inds, "%d", (*ind_pt)); strcat(name, inds); 
      ind_pt++;
    }
    if(!IndexEmpty)
      strcat(name, ")");
 }

 record_vector(name, TRUE, (*v));
#endif /* _VTRACE */
}

/*************************************************************
*                                                            *
*     update_vector is called each time a vector changes     *
*     (i.e., appears in the l.h.s.). It adds the current     *
*     number of nonzeroes among the vector's components,     *
*     timing information and other statistical info to       *
*     the vector's update history.                           *
*                                                            *
*************************************************************/

void update_vector(v, real_time, arity, effect_arity, dstype)
PREC_vectorstruct *v;
int real_time;          /* Time of the update */
int arity;              /* Arity of the r.h.s. */
int effect_arity;       /* Effective arity of the r.h.s. */
int dstype;             /* Data structure used for storing the vector */
{
#ifdef _VTRACE
 pu_struct p;                        /* The vector's update list */

 /* For each time slice of size slice_size ... */
 int min_nz_slice = INT_MAX;         /* Minimum number of nonzeroes */
 int max_nz_slice = -1;              /* Maximum number of nonzeroes */
 int median_nz_slice = 0;            /* Median number of nonzeroes */
 double aver_nz_slice = 0.0;         /* Average number of nonzeroes */
 /* Auxiliaries */
 int i;

 /* Make sure the vector is mapped in the history file */
 account_vector(v);

 /* Compute the global nonzeroes histogram incrementally */
 if(v->nonzeroes < NZ_THRESHOLD) nz_hist[v->nonzeroes]++;
 else                            nz_hist[NZ_THRESHOLD]++;
 /* Compute global min and max incrementally */
 if(v->nonzeroes > max_nz_global) max_nz_global = v->nonzeroes;
 else if(v->nonzeroes < min_nz_global) min_nz_global = v->nonzeroes;

 /* Add a record to the history file (only if global tracking is enabled
    and the vector is tracked or anonymous vectors are tracked) */
 if(global_tracking && (v->is_tracked || track_anonymous)) {
   if(arity != effect_arity) {
     /* Have to give info about effective arity as well 
	(by default equal to arity) */
     fprintf(HF, "e %d\n", effect_arity);
   }
   fprintf(HF, "%d %d %d %d %d\n", v->num_id,
               real_time, v->nonzeroes, dstype, arity);
 }

 /* Do the averaging every slice_size calls if short tracking is enabled */
 if(slice_size) {
   if(call_counter < slice_size)
     nz_array[call_counter++] = v->nonzeroes;
   else {
     for(i=0; i<slice_size; i++) {
        /* Compute average */
	aver_nz_slice += (double)nz_array[i]/(double)slice_size;
     }
     /* Compute min, max, median */
     qsort(nz_array, slice_size, sizeof(int), nz_compare);
     min_nz_slice = nz_array[0];
     max_nz_slice = nz_array[slice_size-1];
     median_nz_slice = nz_array[(int)(slice_size/2)];

     /* Output into the history file */
     fprintf(HF, "s %d %d %d %.2f\n", min_nz_slice, max_nz_slice,
		 median_nz_slice, aver_nz_slice);

     /* Reset */
     call_counter = 0;
   }
 }

#endif /* _VTRACE */
}

/********************************
*                               *
*     Auxiliary functions       *
*                               *
********************************/

int nz_compare(nzp1, nzp2)
void *nzp1;
void *nzp2;
{
#ifdef _VTRACE
 int i1 = *(int*)nzp1;
 int i2 = *(int*)nzp2;
 return (i1 < i2 ? -1 : (i1 > i2 ? 1 : 0));
#endif /* _VTRACE */
}
