/*
  ADIFOR visualization library
  (precision-independent part)

  Vitaly Shmatikov, Summer 1994
*/

#include <limits.h>
#include "advis_pi_include.h"

/*************************************************************
*                                                            *
*     collect_visual is called once for each adxpy call.     *
*     It accepts information about the call (arity of        *
*     the r.h.s., number of zero vectors in the r.h.s.,      *
*     etc.) via its arguments and maintains partial          *
*     statistics in its static variables. If the parameter   *
*     final_call is true, all the accumulated statistics     *
*     is dumped into the file 'advis.dat'.                   *
*                                                            *
*************************************************************/

void collect_visual(arity, effect_arity, no_of_zero_vecs, no_of_zero_mult, no_of_zero_both, lhs_is_in_rhs, old_dstype, new_dstype, final_call)
int arity, effect_arity, no_of_zero_vecs, no_of_zero_mult, no_of_zero_both, old_dstype, new_dstype;
short lhs_is_in_rhs;
short final_call;     /* If final_call is true, output the results */
{
#ifdef _VTRACE

 static long arity_sum = 0;           /* Sum of all arities */
 static long effect_arity_sum = 0;    /* Sum of all effective arities */
 static long dense_add_count = 0;     /* Total # of additions (dense) */
 static long zero_vec_sum = 0;        /* Total # of 0 vectors */
 static long zero_mult_sum = 0;       /* Total # of 0 multipliers */
 static long zero_both_sum = 0;       /* Both vector and multiplier are 0 */
 static long lhs_in_rhs_sum = 0;      /* Total # of times lhs appears in rhs */

 static long no_conv = 0;             /* No conversion */
 static long alloc_conv = 0;          /* "Allocation" conversion */
 static long s1_ss_conv = 0;          /* Various types of conversions */
 static long s1_cs_conv = 0;
 static long ss_s1_conv = 0;
 static long ss_cs_conv = 0;
 static long cs_s1_conv = 0;
 static long cs_ss_conv = 0;

 static int arity_hist[MAXARITY+1];        /* Arity histogram */
 static int eff_arity_hist[MAXARITY+1];    /* Effective arity histogram */

 int i;

 if (!final_call) {

    arity_sum += arity; 
    effect_arity_sum += effect_arity; 
    /* Compute the arity histograms incrementally */
    arity_hist[arity]++;
    eff_arity_hist[effect_arity]++;

    zero_vec_sum += no_of_zero_vecs;
    zero_mult_sum += no_of_zero_mult;
    zero_both_sum += no_of_zero_both;

    if (lhs_is_in_rhs) lhs_in_rhs_sum++;

    dense_add_count += (effect_arity - 1);

    /* Determine the type of conversion (if any) */
    if (old_dstype * new_dstype != 0)
       if (old_dstype == new_dstype) no_conv++;
       else switch (old_dstype) {
	       case S1: if (new_dstype == SS) s1_ss_conv++;
			else if (new_dstype == CS) s1_cs_conv++;
			     else fprintf(stderr, "\nDSTYPE ERROR!\n");
                        break;
	       case SS: if (new_dstype == S1) ss_s1_conv++;
			else if (new_dstype == CS) ss_cs_conv++;
			     else fprintf(stderr, "\nDSTYPE ERROR!\n");
                        break;
	       case CS: if (new_dstype == S1) cs_s1_conv++;
			else if (new_dstype == SS) cs_ss_conv++;
			     else fprintf(stderr, "\nDSTYPE ERROR!\n");
                        break;
               default: fprintf(stderr, "\nDSTYPE ERROR!\n");
			break;
            } 
    else {
       /* old_dstype = 0, i.e. the destination vector was not initialized */
       alloc_conv++;
    }

 }
 else {
   
   /* Final call: dump all information into a file */

   double zero_vec_average = (double)zero_vec_sum/(double)ADXPY_CALLS;
   double zero_mult_average = (double)zero_mult_sum/(double)ADXPY_CALLS;
   double zero_both_average = (double)zero_both_sum/(double)ADXPY_CALLS;
   double lhs_in_rhs_percent = ((double)lhs_in_rhs_sum/(double)ADXPY_CALLS) * 100.0;


   int min_arity = INT_MAX;             /* Minimum arity */
   int max_arity = -1;                  /* Maximum arity */
   int median_arity = 0;                /* Median arity */
   double average_arity;                /* Average arity */
   int arity_calls = 0;

   int min_eff_arity = INT_MAX;         /* Minimum effective arity */
   int max_eff_arity = -1;              /* Maximum effective arity */
   int median_eff_arity = 0;            /* Median effective arity */
   double average_eff_arity;            /* Average effective arity */
   int eff_arity_calls = 0;

   int median_nz_global = 0;            /* Median # of nonzeroes */
   double average_nz_global = 0.0;      /* Average # of nonzeroes */
   int nz_calls = 0;

   /* Auxiliaries */
   short median_computed = FALSE;       /* Median reached already? */
   short min_computed = FALSE;          /* Min computed already? */
   int temp_sum = 0;
   long nz_sum = 0;
   int half_calls;                      /* 1/2 the total # of calls */

   fprintf(DF, "Total number of adxpy calls = %d\n", ADXPY_CALLS);
   fprintf(DF, "Average number of 0 vectors per adxpy call = %.2f\n", zero_vec_average);
   fprintf(DF, "Average number of 0 multipliers per adxpy call = %.2f\n", zero_mult_average);
   fprintf(DF, "Average number of cases where both vector and multiplier are 0 per adxpy call = %.2f\n", zero_both_average);
   fprintf(DF, "Adxpy calls in which the l.h.s appears in the r.h.s. = %.2f%%\n", lhs_in_rhs_percent);

   fprintf(DF, "\nNo conversion: %d calls\n", no_conv);
   fprintf(DF, "\"Allocation\" conversion: %d calls\n", alloc_conv);
   fprintf(DF, "S1-SS conversion: %d calls\n", s1_ss_conv);
   fprintf(DF, "S1-CS conversion: %d calls\n", s1_cs_conv);
   fprintf(DF, "SS-S1 conversion: %d calls\n", ss_s1_conv);
   fprintf(DF, "SS-CS conversion: %d calls\n", ss_cs_conv);
   fprintf(DF, "CS-S1 conversion: %d calls\n", cs_s1_conv);
   fprintf(DF, "CS-SS conversion: %d calls\n", cs_ss_conv);

   fprintf(DF, "\nTotal number of multiplications: %d\n", MULT_COUNT);
   fprintf(DF, "Total number of additions: %d\n", ADD_COUNT);
   fprintf(DF, "Total number of multiplications (dense): pmax * %d\n", effect_arity_sum);
   fprintf(DF, "Total number of additions (dense): pmax * %d\n", dense_add_count);

   /* Output the arity histogram */
   fprintf(DF, "\nArity histogram:\n");
   for(i=0; i<=MAXARITY; i++) 
      arity_calls += arity_hist[i];
   half_calls = arity_calls / 2;
   for(i=0; i<=MAXARITY; i++) {
      fprintf(DF, "%d %d\n", i, arity_hist[i]);
      /* Compute min */
      if(!min_computed && (arity_hist[i] > 0)) {
	min_computed = TRUE; min_arity = i;
      }
      /* Compute max */
      if(arity_hist[i] > 0) max_arity = i;
      /* Compute median */
      temp_sum += arity_hist[i];
      if(!median_computed && (temp_sum > half_calls)) {
	 median_computed = TRUE; median_arity = i;
      }
   }
   /* Compute average */
   average_arity = (double)arity_sum / (double)arity_calls;

   fprintf(DF, "Minimum = %d\n", min_arity);
   fprintf(DF, "Maximum = %d\n", max_arity);
   fprintf(DF, "Median  = %d\n", median_arity);
   fprintf(DF, "Average = %.2f\n", average_arity);

   median_computed = min_computed = FALSE; temp_sum = 0;
   /* Output the effective arity histogram */
   fprintf(DF, "\nEffective arity histogram:\n");
   for(i=0; i<=MAXARITY; i++) 
      eff_arity_calls += eff_arity_hist[i];
   half_calls = eff_arity_calls / 2;
   for(i=0; i<=MAXARITY; i++) {
      fprintf(DF, "%d %d\n", i, eff_arity_hist[i]);
      /* Compute min */
      if(!min_computed && (eff_arity_hist[i] > 0)) {
	min_computed = TRUE; min_eff_arity = i;
      }
      /* Compute max */
      if(eff_arity_hist[i] > 0) max_eff_arity = i;
      /* Compute median */
      temp_sum += eff_arity_hist[i];
      if(!median_computed && (temp_sum > half_calls)) {
	 median_computed = TRUE; median_eff_arity = i;
      }
   }
   /* Compute average */
   average_eff_arity = (double)effect_arity_sum / (double)eff_arity_calls;

   fprintf(DF, "Minimum = %d\n", min_eff_arity);
   fprintf(DF, "Maximum = %d\n", max_eff_arity);
   fprintf(DF, "Median  = %d\n", median_eff_arity);
   fprintf(DF, "Average = %.2f\n", average_eff_arity);

   median_computed = min_computed = FALSE; temp_sum = 0; nz_sum = 0;
   /* Output the nonzeroes histogram */
   fprintf(DF, "\nNonzeroes histogram:\n");
   for(i=0; i<=NZ_THRESHOLD; i++) {
      nz_calls += nz_hist[i];
      nz_sum += nz_hist[i] * i;
   }
   half_calls = nz_calls / 2;
   for(i=0; i<=NZ_THRESHOLD; i++) {
      fprintf(DF, "%d %d\n", i, nz_hist[i]);
      /* Compute median */
      temp_sum += nz_hist[i];
      if(!median_computed && (temp_sum > half_calls)) {
	 median_computed = TRUE; median_nz_global = i;
      }
   }
   /* Compute average */
   average_nz_global = (double)nz_sum / (double)nz_calls;

   fprintf(DF, "Minimum = %d\n", min_nz_global);
   fprintf(DF, "Maximum = %d\n", max_nz_global);
   fprintf(DF, "Median  = %d\n", median_nz_global);
   fprintf(DF, "Average = %.2f\n", average_nz_global);
 }
#endif /* _VTRACE */
}

/*************************************************************
*                                                            *
*     visual_mark puts a numbered mark into the history      *
*     file                                                   *
*                                                            *
*************************************************************/

void visual_mark(mark)
int *mark;
{
#ifdef _VTRACE
 fprintf(HF, "m %d\n", (*mark));
#endif /* _VTRACE */
}
