/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/


#include  <stdlib.h>
#include "gaxpy_pi_macros_and_data_structures.h"
#include "gaxpy_pi_global_variables.h"


/****************************************************************************
 *                                                                          *
 *  Variables Global to routines in this file.                              *
 *                                                                          *
 ****************************************************************************/

static heapstruct *heapstore[MAXARITY];      /* Array of addresses of the   *
					      * heap nodes.  Note: unlike   *
					      * buckets and vectors, the    *
					      * total # of heap nodes       *
					      * needed for each saxpy is <= *
					      * to the arity which is known *
					      * apriori, and thus we don't  *
					      * need a growable data        *
					      * structure to represent the  *
					      * heapstore.                  */

/****************************************************************************
 *                                                                          *
 *  allocateheapstore()                                                     *
 *                                                                          *
 *  allocateheapstore allocates all the memory for the heapstore.  It is    *
 *  called only once.                                                       *
 *                                                                          *
 ****************************************************************************/

void allocateheapstore()
{
  int k;
  
  for (k=0;k<MAXARITY;k++) 
    if((heapstore[k] =  (heapstruct *) palloc(sizeof(heapstruct))) == NULL) {
      printf("allocateheapstore error: malloc returned NULL. k=%d\n",k);
      abort();
    } 
}

/****************************************************************************
 *                                                                          *
 *  heapinitialize(vectorindex,subscriptindex)                              *
 *                                                                          *
 *     vectorindex     Identifies the vector to which the nonzero belongs.  *
 *                                                                          *
 *     subscriptindex  Array index of the nonzero in the non-sparse vector. *
 *                                                                          *
 *     heapinitialize assigns to the heaproot the first nonzero in the      *
 *     first vector operand in the rhs of the saxpy, by putting the vector  *
 *     and subscript indices in the heaproot.                               *
 *                                                                          *
 ****************************************************************************/

void heapinitialize(vectorindex,rangebeginindex,rangeendindex)
int vectorindex,rangebeginindex,rangeendindex;

{
  heapstoreptr = 0;
  heaproot = heapstore[heapstoreptr];
  heaproot->vectorindex = vectorindex;

  heaproot->rangebeginindex = rangebeginindex;
  heaproot->rangeendindex = rangeendindex;

  heaproot->left = NULL;
  heaproot->right = NULL;
}

/****************************************************************************
 *                                                                          *
 *  heapinsert(vectorindex,subscriptindex)                                  *
 *                                                                          *
 *     vectorindex     Identifies the vector to which the nonzero belongs.  *
 *                     It's equal to the order of the vector in the rhs of  *
 *                     the saxpy.                                           *
 *                                                                          *
 *     subscriptindex  Array index of the nonzero in the non-sparse vector. *
 *                     It's used as the key for the heapsort.               *
 *                                                                          *
 *     heapinsert inserts a new node into the binary tree in O(log(arity)). *
 *     heapinsert always starts out with the heaproot having been assigned  *
 *     by heapinitialize.  It gets a new heap node from the heapstore and   *
 *     sets the values of the newnode's non-pointer fields to the function  *
 *     arguments, vectorindex and subscriptindex.  It then uses heapptr to  *
 *     search for the "right" place in the heap for the newnode.  heapptr   *
 *     starts out pointing to the root.  Using subscriptindex as key, it    *
 *     descends down the tree WHILE the key value of the node N, it's       *
 *     pointing to is smaller than the newnode's key.  Within the WHILE     *
 *     loop, if a node N is encountered with a free child, the new node     *
 *     the child of node N.  If, however, we descend to the "right" place   *
 *     in the tree and there are no free children (i.e., node N's key is >= *
 *     the newnode's key), node N and the newnode "switch places and we     *
 *     reenter the loop.  The following 2 heapinsert snapshot scenarios     *
 *     show the  insertion of newnodes into trees (we show each node        *
 *     containing only its key):                                            *
 *                                                                          *
 *     scenario 1:                                                          *
 *                                                                          *
 *     newnode --> <5>          newnode --> <6>                             *
 *                                                                          *
 *     heapptr heaproot         heapptr heaproot       heapptr    heaproot  *
 *        |       |                |       |              |          |      *
 *        |       V                |       V              |          V      *
 *        +----> <2>               |      <2>             |         <2>     *
 *               / \               |      / \             |         / \     *
 *             <6> <4>     ==>     +--> <5> <4>   ==>     |       <5> <4>   *
 *             / \                      / \               |       / \       *
 *           <8> <9>                  <8> <9>             |     <6> <9>     *
 *                                                        |     /           *
 *                                                        +-> <8>           *
 *                                                                          *
 *     scenario 2:                                                          *
 *                                                                          *
 *     newnode --> <2>    newnode --> <4>     newnode --> <8>               *
 *                                                                          *
 *     heapptr heaproot   heapptr heaproot    heapptr heaproot     heaproot *
 *        |       |          |       |           |       |            |     *
 *        |       V          |       V           |       V            V     *
 *        +----> <4>   ==>   +----> <2>   ==>    |      <2>    ==>   <2>    *
 *               / \                / \          |      / \          / \    *
 *             <8> <6>            <8> <6>        +--> <4> <6>      <4> <6>  *
 *                                                                     /    *
 *                                                       heapptr --> <8>    *
 *                                                                          *
 ****************************************************************************/

void heapinsert(vectorindex,rangebeginindex,rangeendindex)
int vectorindex,rangebeginindex,rangeendindex;

{
  heapstruct *heapnewnode;                   /* nsia.                       */

  heapstruct *heapptr;                       /* Address of "current" node   *
					      * during a pass of the tree.  */

  heapstruct *heapptrparent;                 /* nsia.                       */

  heapstruct *heapptrleftchild;              /* nsia.                       */

  heapstruct *heapptrrightchild;             /* nsia.                       */

  heapstruct *tempheapptr;                   /* nsia.                       */

  heapnewnode = heapstore[++heapstoreptr];
  heapnewnode->vectorindex = vectorindex;

  heapnewnode->rangebeginindex = rangebeginindex;
  heapnewnode->rangeendindex = rangeendindex;

  heapptrparent = NULL;
  heapptr = heaproot;

  while (heapptr !=NULL) {

    while (heapptr !=NULL &&
	   heapnewnode->rangebeginindex >= heapptr->rangebeginindex) {
      heapptrleftchild = heapptr->left;
      heapptrrightchild = heapptr->right;
      if (heapptr->left !=NULL && heapptr->right !=NULL) {

	if (heapptrleftchild->rangebeginindex >=
	    heapptrrightchild->rangebeginindex) {
	  heapptrparent = heapptr;
	  heapptr = heapptrleftchild;
	}
	else {
	  heapptrparent = heapptr;
	  heapptr = heapptrrightchild;
	}
      }
      else if (heapptr->left == NULL) {
	heapptr->left = heapnewnode;
	heapnewnode->left = NULL;
	heapnewnode->right = NULL;
	heapptr = NULL;
      }
      else {
	heapptr->right = heapnewnode;
	heapnewnode->left = NULL;
	heapnewnode->right = NULL;
	heapptr = NULL;
      }
    }
    if (heapptr !=NULL) {
      heapnewnode->left = heapptr->left;
      heapnewnode->right = heapptr->right;
      tempheapptr = heapptr;
      if (heapptr == heaproot)
	heaproot = heapnewnode;
      else
	if (heapptrparent->left == heapptr)
	  heapptrparent->left = heapnewnode;
	else
	  heapptrparent->right = heapnewnode;
      heapptr = heapnewnode;
      heapnewnode = tempheapptr;
    }
  }
}

/****************************************************************************
 *                                                                          *
 *  heapop(operationlabel,vectorindex,subscriptindex)                       *
 *                                                                          *
 *     operationlabel   == REPLACE means:  replace heaproot with the new    *
 *                                         heap entry, as defined by the    *
 *                                         next 2 arguments.                *
 *                      == DELETE means :  delete heaproot.  The next two   *
 *                                         arguments are dummy values.      *
 *                                                                          *
 *     vectorindex     If operationlabel == REPLACE, then vectorindex means *
 *                     the same as it did in heapinsert; else it's a dummy  *
 *                     value.                                               *
 *                                                                          *
 *     subscriptindex  If operationlabel == REPLACE, then subscriptindex    *
 *                     means the same as it did in heapinsert; else it's a  *
 *                     very large value meant to ensure the heaproot node   *
 *                     "sinks".                                             *
 *                                                                          *
 *     heapop traverses the tree in O(log(arity)) and depending on the      *
 *     operationlabel, either replaces the root with the new entry or       *
 *     deletes the root.  These two operations are implemented as one       *
 *     function since they share the same basic control flow.  We will      *
 *     first describe the replace operation and then describe how the       *
 *     delete operation differs from it.                                    *
 *                                                                          *
 *     heapop(REPLACE,...) is called when the nonzero in the heaproot is    *
 *     consumed in the saxpy operation and needs to be replaced in the heap *
 *     by the next nonzero from the same vector.  heapptr is assigned the   *
 *     address of the heaproot.  The non-pointer fields in heapptr (i.e.,   *
 *     heaproot) are replaced by the function arguments, vectorindex and    *
 *     subscriptindex.  After this initialization, heapptr remains pointing *
 *     at the same node, throughout the rest of the function.  The WHILE    *
 *     loop is active while we haven't found the right place in the tree    *
 *     for the new key in heapptr, i.e., while for at least one of the      *
 *     children, key_heapptr > key_child.  Inside the WHILE, we switch the  *
 *     placement within the tree of heapptr and  one of the children for    *
 *     which the above condition was satisfied.  This is basically done in  *
 *     the following manner:                                                *
 *        IF heapptr needs to switch places with leftchild                  *
 *           { switch the places of heapptr and leftchild }                 *
 *        ELSE                                                              *
 *           { switch the places of heapptr and rightchild }                *
 *     The switching of places is done through pointer reassignments.       *
 *                                                                          *
 *     heapop(DELETE,...) is called when there are no more nonzeroes left   *
 *     in the vector whose nonzero was most recently consumed in the saxpy. *
 *     The subscriptindex is given a value (largest 16-bit positive         *
 *     integer) greater than any vector subscript, thus ensuring that the   *
 *     heapptr node "sinks" or "bubbles" down to the bottom of the tree     *
 *     (i.e., becomes a leaf), and is then cut off from the tree.           *
 *                                                                          *
 ****************************************************************************/

void heapop(operationlabel,vectorindex,rangebeginindex,rangeendindex)
int operationlabel,vectorindex,rangebeginindex,rangeendindex;

{
  heapstruct *heapptr;                       /* Initial address of the      *
          				      * heaproot node.  heapptr     *
					      * always points to the same   *
					      * places.                     */

  heapstruct *heapptrparent;                 /* nsia.                       */

  heapstruct *heapptrleftchild;              /* nsia.                       */

  heapstruct *heapptrrightchild;             /* nsia.                       */

  heapstruct *tempheapptr;                   /* nsia.                       */
  
  heapptrparent = NULL;
  heapptr = heaproot;
  heapptr->vectorindex = vectorindex;

  heapptr->rangebeginindex = rangebeginindex;
  heapptr->rangeendindex = rangeendindex;

  heapptrleftchild = heapptr->left;
  heapptrrightchild = heapptr->right;

  while ((heapptrleftchild !=NULL &&
	  heapptr->rangebeginindex > heapptrleftchild->rangebeginindex) ||
	 (heapptrrightchild !=NULL &&
	  heapptr->rangebeginindex > heapptrrightchild->rangebeginindex)) {
    if (((heapptrleftchild !=NULL && heapptrrightchild !=NULL) &&
	 (heapptrleftchild->rangebeginindex <=
	  heapptrrightchild->rangebeginindex)) ||
	(heapptrleftchild !=NULL && heapptrrightchild ==NULL)) {
      tempheapptr = heapptrrightchild;
      heapptr->left = heapptrleftchild->left;
      heapptr->right = heapptrleftchild->right;
      heapptrleftchild->left = heapptr;
      heapptrleftchild->right = tempheapptr;
      if (heapptr == heaproot) {
	heaproot = heapptrleftchild;
      }
      else if (heapptrparent->left == heapptr) {
	heapptrparent->left = heapptrleftchild;
      }
      else if (heapptrparent->right == heapptr) {
	heapptrparent->right = heapptrleftchild;
      }
      
#ifdef _DEBUG
      else {
	printf("heapop error: leftchild=%p, rightchild=%p, child=%p\n",
	       heapptrparent->left,heapptrparent->right,heapptr);
	abort();
      }
#endif

      heapptrparent = heapptrleftchild;
    }

    else if (((heapptrleftchild !=NULL && heapptrrightchild !=NULL) &&
	      (heapptrleftchild->rangebeginindex > 
	       heapptrrightchild->rangebeginindex)) ||
	     (heapptrleftchild ==NULL && heapptrrightchild !=NULL)) {
      tempheapptr = heapptrleftchild;
      heapptr->right = heapptrrightchild->right;
      heapptr->left = heapptrrightchild->left;
      heapptrrightchild->right = heapptr;
      heapptrrightchild->left = tempheapptr;
      if (heapptr == heaproot) {
	heaproot = heapptrrightchild;
      }
      else if (heapptrparent->left == heapptr) {
	heapptrparent->left = heapptrrightchild;
      }
      else if (heapptrparent->right == heapptr) {
	heapptrparent->right = heapptrrightchild;
      }
      
#ifdef _DEBUG
      else {
	printf("heapop error: leftchild=%p, rightchild=%p, child=%p\n",
	       heapptrparent->left,heapptrparent->right,heapptr);
	abort();
      }
#endif
      
      heapptrparent = heapptrrightchild;
    }
    heapptrleftchild = heapptr->left;
    heapptrrightchild = heapptr->right;
  }
  
  if (operationlabel == DELETE) {
    if (heapptrparent !=NULL) {
      if (heapptrparent->left == heapptr)
	heapptrparent->left = NULL;
      else if (heapptrparent->right == heapptr)
	heapptrparent->right = NULL;
      
#ifdef _DEBUG
      else {
	printf("heapop error: leftchild=%p, rightchild=%p, child=%p\n",
	       heapptrparent->left,heapptrparent->right,heapptr);
	abort();
      }
#endif
      
    }
    else
      heaproot = NULL;
  }
}

