/*------------------------------------------------------------------*/
/* (C) COPYRIGHT 1994 University of Chicago                         */
/*                                                                  */
/* THIS PROGRAM DISCLOSES MATERIAL PROTECTABLE UNDER COPYRIGHT      */
/* LAWS OF THE UNITED STATES. USE IS PERMITTED WITHOUT FEE FOR      */
/* EDUCATION AND NON-PROFIT RESEARCH PURPOSES SUBJECT TO THE        */
/* TERMS OF A LICENSE DISTRIBUTED WITH THE ADIFOR 2.0 SOFTWARE      */
/* AND AVAILABLE AT:                                                */
/* http://www.mcs.anl.gov/autodiff/adifor_license.html, or          */
/* http://www.cs.rice.edu/fortran-tools/ADIFOR/adifor_license.html. */
/* ANY ENTITY DESIRING PERMISSION TO INCORPORATE THIS SOFTWARE      */
/* OR A WORK BASED ON THE SOFTWARE INTO COMMERCIAL PRODUCTS OR      */
/* OTHERWISE USE IT FOR COMMERCIAL PURPOSES SHOULD CONTACT:         */
/*                                                                  */
/* Paul Hovland, Mathematics and Computer Science Division,    */
/* Argonne National Laboratory, 9700 S. Cass Avenue, Argonne IL     */
/* 60439, hovland@mcs.anl.gov.                                      */
/*                                                                  */
/* Alan Carle, Center for Research on Parallel Computation, Rice    */
/* University MS 41, 6100 S. Main Street, Houston, TX 77251-1892,   */
/* carle@cs.rice.edu.                                               */
/*------------------------------------------------------------------*/

#ifndef MULTITYPE_H
#define MULTITYPE_H 1

/* math.h: We need the sqrt() function for complex X_ABS() */
#include <math.h>
#include "complex-typedefs.h"

/*
   This header will appropriately define addition, subtraction, 
   and multiplication for single, double, complex, and double complex.

   We have to assume that each X_ASSIGN, X_ADD, X_SUBT, X_MULT, and
   X_DIV occurs on a line by itself. (Complex expressions are meaningless.)
*/

#ifdef SINGLE

#define S
#define REAL_ARITHMETIC
#define FLOATING_DECLARE float
#define BASE_FLOATING_DECLARE float

#endif /* SINGLE */


#ifdef DOUBLE

#define D
#define REAL_ARITHMETIC
#define FLOATING_DECLARE double
#define BASE_FLOATING_DECLARE double

#endif /* DOUBLE */


#ifdef COMPLEX 

#define C
#define COMPLEX_ARITHMETIC
#define FLOATING_DECLARE complex
#define BASE_FLOATING_DECLARE float

#endif /* COMPLEX */


#ifdef DOUBLE_COMPLEX

#define Z
#define COMPLEX_ARITHMETIC
#define FLOATING_DECLARE double_complex
#define BASE_FLOATING_DECLARE double

#endif /* DOUBLE_COMPLEX */

/*
   Now the arithmetic.
*/


#ifdef REAL_ARITHMETIC
/* These must be statements */
#define X_ASSIGN(y,x) (y) = (x)
#define X_ADD(y,a,b) (y) = (a) + (b)
#define X_SUBT(y,a,b) (y) = (a) - (b)
#define X_MULT(y,a,b) (y) = (a) * (b)
#define X_DIV(y,a,b) (y) = (a) / (b)

#define X_PLUS(y,a,b) X_ADD(y,a,b)
#define X_PLUS_EQUALS(y,a) (y) += (a)

#define X_PLUS_MULT(y,a,b) (y) += (a) * (b)

#define X_PUT_REAL(y,x) (y) = (x)
#define X_GET_AND_CAST_REAL(y,x,cast) (y) = (cast)(x)
#define X_ZERO(y) X_PUT_REAL(y,0)

/* These are expressions */
#define X_ABS(y) ( ((y) > 0) ? (y) : -(y) )
#define X_EQUAL(x,y) ( (x) == (y) )
#define X_NE(x,y) ( (x) != (y) )
#define X_NEQ(x,y) X_NE(x,y)
#define X_EQ_REAL(x,y) ( x == y)
#define X_IS_ZERO(x) X_EQ_REAL(x,0)
#define X_IS_ONE(x) X_EQ_REAL(x,1)

/* These are stylized print statement(s) */
#define X_PRINT_D_E(a,b) \
     ( printf(" %8d    %13.6e\n", (a), (b) ) )
#define X_FPRINT_D_E(f,a,b) \
     ( fprintf((f)," %8d    %13.6e\n", (a), (b) ) )

#endif /* REAL_ARITHMETIC */


#ifdef COMPLEX_ARITHMETIC

/* These must be statements */

#define X_ASSIGN(y,x) \
     (y).real = (x).real ; \
     (y).imag = (x).imag;

#define X_ADD(y,a,b) \
     (y).real = (a).real + (b).real ; \
     (y).imag = (a).imag + (b).imag ; 

#define X_SUBT(y,a,b) \
     (y).real = (a).real - (b).real ; \
     (y).imag = (a).imag - (b).imag ; 

#define _X_MULT_REAL(a,b) \
     (a).real * (b).real - (a).imag * (b).imag

#define _X_MULT_IMAG(a,b) \
     (a).real * (b).imag + (a).imag * (b).real ;

#define X_MULT(y,a,b) \
     (y).real = _X_MULT_REAL(a,b); \
     (y).imag = _X_MULT_IMAG(a,b);

#define X_DIV(y,a,b) \
{ \
     BASE_FLOATING_DECLARE mulTITYPE_quick_div_tmp; \
     mulTITYPE_quick_div_tmp = ((b).real * (b).real + (b).imag * (b).imag); \
     (y).real = (((a).real * (b).real + (a).imag * (b).imag) \
		 / mulTITYPE_quick_div_tmp) ;  \
     (y).imag = (((a).imag * (b).real - (a).real * (b).imag) \
		 / mulTITYPE_quick_div_tmp) ;  \
}

#define X_PLUS(y,a,b) X_ADD(y,a,b)

#define X_PLUS_EQUALS(y,a) \
     (y).real += (a).real ; \
     (y).imag += (a).imag ; 

#define X_PLUS_MULT(y,a,b) \
     (y).real += _X_MULT_REAL(a,b); \
     (y).imag += _X_MULT_IMAG(a,b);

#define X_PUT_REAL(y,x) (y).real = x ; (y).imag = 0;
#define XY_PUT_AND_CAST(z,x,y,cast) (z).real = (cast)(x); \
                                    (z).imag = (cast)(y);
#define XY_PUT_AND_CAST_PLUS_EQUALS(z,x,y,cast) (z).real += (cast)(x); \
                                                (z).imag += (cast)(y);

#define X_GET_AND_CAST_REAL(y,x,cast) (y) = (cast)(x).real;
#define X_GET_AND_CAST_IMAG(y,x,cast) (y) = (cast)(x).imag;
#define X_CONJUGATE(y,x) (y).real = (x).real; \
                         (y).imag = -(x).imag;
#define X_ZERO(y) X_PUT_REAL(y,0)

/* These are expressions */
#define X_ABS(y) ( sqrt(((y).real * (y).real) + ((y).imag * (y).imag)) )
#define X_EQUAL(x,y) ( ((x).real == (y).real) && ((x).imag == (y).imag) )
#define X_NE(x,y) ( ! X_EQUAL(x,y))
#define X_NEQ(x,y) X_NE(x,y)
#define X_EQ_REAL(x,y) ( ((x).real == y) && ((x).imag == 0) )
#define X_IS_ZERO(x) X_EQ_REAL(x,0)
#define X_IS_ONE(x) X_EQ_REAL(x,1)

/* These are stylized print statement(s) */
#define X_PRINT_D_E(a,b) \
     ( printf(" %8d    (%13.6e, %13.6e)\n", (a), (b).real, (b).imag ) )
#define X_FPRINT_D_E(f,a,b) \
     ( fprintf((f)," %8d    (%13.6e, %13.6e)\n", (a), (b).real, (b).imag ) )

/* These are statements needed for real-complex arithmetic */

#define X_IMAG(y,x) (y).imag = x;

#endif /* COMPLEX_ARITHMETIC */

#endif /* ndef MULTITYPE_H */
